#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "expr.h"
#include "widget.h"
#include "strbuf.h"
#include "eval.h"
#include "library.h"
#include "ptrbuf.h"
#include "gen.h"

static char *dotless_name(char *str, char *pfx)
{
  char *Z;
  strbuf_t *SB;
  SB= strbuf_init();
  strbuf_puts(SB, pfx);
  strbuf_puts(SB, "_");
  while(1)
  {
    Z= strchr(str,'.');
    if (Z)
    {
      if (Z!=str)
        strbuf_puts_l(SB, str, Z-str);
      strbuf_puts(SB,"_XO_");
      str= Z+1;
      if (!str[0]) break;
    }
    else
    {
      strbuf_puts(SB,str);
      break;
    }
  }
  return strbuf_collect(SB);
}

static char *widget_layout_fname(Uscreen_t *scr)
{
  return dotless_name(scr->ident, "LAYOWID");
}

static char *screen_layout_fname(Uscreen_t *scr)
{
  return dotless_name(scr->ident,"LAYOSCR");
} 

static void gen_widget_layout(Uscreen_t *scr,FILE* f)
{
  char *code;
  size_t nunk;
  unknown_t **unk;
  int N;
  Uwidget_t *W;

  fprintf(f,"static void %s(int WIDTH,int HEIGHT,int *L) {\n",
                  widget_layout_fname(scr));
  code= eval_widget_layout(scr, &nunk, &unk);
  if (nunk)
  { 
     int i;
     fprintf(f,"%s",code);
     fprintf(f,"#error \"there are unknowns, eval this first\"\n");
     for(i=0;i<nunk;i++)
      fprintf(f,"// %s= %s \n", unk[i]->target, unk[i]->str);
     fprintf(f,"}\n");
     return ;
  }
  fprintf(f,"%s",code);
  free(code);

  for(N=0,W=scr->widgets;W;W=W->next,N++)
  {
    fprintf(f, "  L[%d]= CO%s_left; L[%d]= CO%s_top;\n" 
               "  L[%d]= CO%s_width; L[%d]= CO%s_height;\n", 
                     N*4, W->ident, N*4+1, W->ident,
                     N*4+2, W->ident, N*4+3, W->ident);
  }
  fprintf(f,"}\n");
}

static void gen_screen_dim(Uscreen_t *scr, FILE *f)
{
  fprintf(f,"static void %s(int *L,int WIDTH,int HEIGHT) {\n",
                       screen_layout_fname(scr));
  fprintf(f,"%s", eval_screen_dim(scr));
  fprintf(f,"  L[0]= COleft;\n");
  fprintf(f,"  L[1]= COtop;\n");
  fprintf(f,"  L[2]= COwidth;\n");
  fprintf(f,"  L[3]= COheight;\n");
  fprintf(f,"}\n");
}

static void gen_widget_names(FILE *f)
{
  Uscreen_t *scr;
  Uwidget_t *W;
  Uscreen_t **SS; int N;
  int i;
  SS= get_all_screens(&N);

  fprintf(f,"enum { _XX_WIDGETS_START=GFXEV_LAST+GFXRESERVED,\n ");
  fprintf(f," /* WIDGET NAMES BELOW */\n");
  char buf[2000];
  buf[0]= 0;
  for(i=0;i<N;i++)
  {
   scr= SS[i];
   for(W=scr->widgets;W;W=W->next)
   {
     fprintf(f,"  %s%s,\n", W->ident,buf);
     if (!strcmp(W->klass->ident,"PixGrid"))
     {
       sprintf(buf,"=%s+GFXEV_LAST",W->ident);
     }
     else buf[0]= 0;
   }
  }
  fprintf(f,"\n _XX_WIDGETS_END };\n");
  free(SS);
}

static void declare_widget_list(Uwidget_t **WL, int n, FILE *f)
{
  int i;
  fprintf(f," static int SWID[]= { ");
  for(i=0;i<n;i++)
    fprintf(f,"    %s,\n", WL[i]->ident);
  fprintf(f,"};\n");
}

static Uwidget_t** get_widget_list_breadthfirst(Uscreen_t *scr,int *Rn)
{
  // It's important to get the list in breadthfirst fashion,
  // this is because we want to calculate the size&pos for the
  // parent widgets first, and then the child widgets. 
  ptrbuf_t *queue;
  ptrbuf_t *result;
  Uwidget_t *W;
  int i;
  size_t Nww; Uwidget_t **ww;
  int nresult; // #widgets in result so far

  result= ptrbuf_new(20); // holds the widgets
  queue= ptrbuf_new(20); // holds the screens
  nresult= 0;

  ptrbuf_put(queue, scr);
  while((scr=ptrbuf_get(queue)))
  {
    scr->windex_start= nresult;
    for(W=scr->widgets;W;W=W->next)
    {
      ptrbuf_put(result, W);
      nresult++;
      scr->windex_end= nresult;
      for(i=0;i<W->n_subs;i++) ptrbuf_put(queue, W->subs[i]);
    }
  }
  ptrbuf_free(queue,NULL,NULL);
  ptrbuf_free(result, (void***) &ww, &Nww);
  *Rn= Nww;
  return ww;
}

static void Gcreate_widget(Uwidget_t *W,FILE *f)
{
  (*W->klass->create)(W,f);
  fprintf(f, "  widgets[%s]->number= %s;\n", W->ident, W->ident);
}

static void Ginitialmove_widget(Uwidget_t *W,FILE *f,int N)
{
  fprintf(f,"  widget_move(widgets[%s], LC[%d], LC[%d],LC[%d],LC[%d]);\n",
        W->ident, N*4,N*4+1,N*4+2,N*4+3);
}

static void Gshow_widget(Uwidget_t *W,FILE *f)
{
    fprintf(f,"  widget_show(widgets[%s], 1);\n", W->ident);
}

static void gen_screen_func(Uscreen_t *scr,FILE *f,int mode)
{
  Uwidget_t *W;
  int i,j;
  Uscreen_t *S;

  Uwidget_t **ww; int Nww; // this is the list of ALL widgets,
                           // including those in the sub widgets.

  fprintf(f,"void gfx_%s()",scr->ident);
  if (mode=='H') { fprintf(f,";\n"); return ; }
  fprintf(f,"{\n");

  ww= get_widget_list_breadthfirst(scr,&Nww);

  declare_widget_list(ww, Nww,f);
  fprintf(f,"  int i;\n");
  fprintf(f,"  int LC[%d]; /*widget coords */\n",Nww*4);
  fprintf(f,"  int LL[4]; /* screen coords */\n");
  fprintf(f,"  %s(LL,x_window_width,x_window_height);\n",
                   screen_layout_fname(scr));
  fprintf(f,"  enter_screen(%d, SWID,LL);\n",Nww);

                          // we create everything here, including
                          // the widgets embedded in panels.
  for(i=0;i<Nww;i++) Gcreate_widget(ww[i],f);
  // FIXED: the widgets need to be positioned according to the
  //        screen dimensions, not window dimensions
  // FIXME: remove this comment
  //fprintf(f,"  %s(x_window_width,x_window_height,LC);\n", 
  //             widget_layout_fname(scr));
  fprintf(f,"  %s(LL[2],LL[3],LC);\n", widget_layout_fname(scr));

  // Now we need to layout the subscreens.
  // we do that using the breadth-first widget list, which gives us
  // a breadth-first screen list.
  for(i=0;i<Nww;i++)
  {
    W= ww[i];
    if (!W->n_subs) continue;
    for(j=0;j<W->n_subs;j++)
    {
      S= W->subs[j];
      fprintf(f,"  %s(LC[%d],LC[%d], LC+%d);\n",
               widget_layout_fname(S), 4*i+2, 4*i+3, 4*S->windex_start);
      fprintf(f,"  for(i=%d;i<%d;i++) { LC[4*i]+= LC[%d]; LC[4*i+1]+= LC[%d];}\n",
               S->windex_start, S->windex_end, 4*i, 4*i+1);
    }
  }

  fprintf(f,"  for(i=0;i<%d;i++) { LC[4*i]+= LL[0]; LC[4*i+1]+= LL[1]; }\n",Nww);
  for(i=0;i<Nww;i++) Ginitialmove_widget(ww[i],f,i);
                            // panels: the list of shown widgets depend on
                            // the active subscreen of the panel, so this
                            // will work.
  for(W=scr->widgets;W;W=W->next) Gshow_widget(W,f);
  fprintf(f,"  gfx_update_all();\n");
  fprintf(f,"}\n");
}

static char *replace_suffix(char *str,char *sfx)
{
  int i;
  str= strdup(str);
  for(i=strlen(str)-1;i>0;i--)
     if (str[i]=='.') { strcpy(str+i,sfx); break; }
  return  str;
}


static void gen_colorimg_resource(FILE *f,resource_t *R)
{
  int i;
  int L;
  fprintf(f,"RES_COLORIMG(%s,\"User resource\")",R->ident);
  L= R->length/4;
  for(i=0;i<L;i++)
  {
    if (i%5==0) fprintf(f,"\n");
    fprintf(f,"0x%08x,", *(uint32_t*) (R->data+4*i));
  }
  fprintf(f,"\nRES_END\n");
}

static void gen_resources(FILE *f)
{
  resource_t *R;
  for(R=resources;R;R=R->next)
  {
   switch(R->type)
   {
   case TYP_COLORIMG:
    gen_colorimg_resource(f,R);
    break;
   }
  }
}

static void gen_resource_iface(FILE *f)
{
  resource_t *R;
  for(R=resources;R;R=R->next)
  {
   switch(R->type)
   {
   case TYP_COLORIMG:
    fprintf(f, "extern unsigned int %s[];\n", R->ident);
    break;
   }
  }
}

static void generate_header(FILE *out)
{
  const char *Chdr;
  int i,N;
  Uscreen_t **SS,*S;

  get_library_code(&Chdr, NULL);

  fwrite(Chdr, 1, strlen(Chdr), out);

  gen_widget_names(out);
  SS= get_all_screens(&N);

  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (S->type) continue;
    gen_screen_func(S, out,'H'); 
  }
  free(SS);
  gen_resource_iface(out);
}

const char *print_to_mark(const char *code, const char *mark, FILE *f)
{
  const char *z;
  z= strstr(code,mark);
  if (z)
    { while(*z!=0 && *z!='\n') z++;
      if (*z=='\n') z++; }
  else
    z= code+strlen(code);
  if (f)
    fwrite(code, 1, z-code,f);
  return z;
}


static void generate_source_file(char *fn)
{
  FILE *out;
  const char *Ccode;
  int i,N;
  Uscreen_t **SS,*S;

  out= fopen(fn, "w");
  if (!out) return ; // FIXME: give error message

  if (getenv("GFXGEN_DBG")) 
    Ccode= "";
  else
    get_library_code(NULL, &Ccode);

  Ccode= print_to_mark(Ccode, "/** INCLUDE DIRECTIVE BEGIN **/", out);
  generate_header(out);
  Ccode= print_to_mark(Ccode, "/** INCLUDE DIRECTIVE END **/", NULL);
  Ccode= print_to_mark(Ccode, "/** USER RESOURCES **/", out);
  gen_resources(out);
  fwrite(Ccode, 1, strlen(Ccode), out);

  SS= get_all_screens(&N);

  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (!S->type) continue;
    gen_widget_layout(S, out);
  }
 
  for(i=0;i<N;i++)
  {
    S= SS[i];
    if (S->type) continue;
    gen_widget_layout(S, out);
    gen_screen_dim(S, out);
    gen_screen_func(S, out,'C'); 
  }
  free(SS);

  fclose(out);
}

void generate_code(char *hfilenm,char *cfilenm)
{
  FILE *out;
  if (cfilenm) generate_source_file(cfilenm);
  if (!hfilenm) return ;
  out= fopen(hfilenm,"w");
  if (!out) return ;
  generate_header(out);
  fclose(out);
}

