#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "expr.h"
#include "widget.h"
#ifndef KLASSES_NO_GFX
#include "gfx.h"
#endif

#ifdef KLASSES_NO_GFX
#define EDITFUNC(x) NULL
#else
#define EDITFUNC(x) x
#endif

static char *get_field(Uwidget_t *W, char *name, char *def)
{
  field_t *F;
  char *D;
  int i;
  Uklass_t *K;

  for(F=W->fields;F;F=F->next)
    if (!strcmp(F->name, name))
    {
      D= F->data;
      if (D && D[0]) return D;
      break;
    }

  K= W->klass;
  for(i=0;i<K->nfields;i++)
    if (!strcmp(K->fields[i].name,name))
    {
      D= K->fields[i].data;
      if (D && D[0]) return D;
      break;
    }

  return def;
}

static char *nustr(char *x)
{
  if (x) return x;
  else return "NULL";
}

static char *Cstr(char *x)
{
  char *R;
  int L;
  int i,j;
  char *legal= " abcdefghijklmnopqrstuvwxyz!@#$%^&*()_+=-[]{};:,./<>?`~"
               "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
  // char *hex= "0123456789abcdef";
  if (!x) return "NULL";
  L= strlen(x);
  R= malloc(4*L+2+1);
  j= 0;
  R[j++]= '"';
  for(i=0;x[i];i++)
   if (strchr(legal,x[i])) R[j++]= x[i];
   else { sprintf(R+j,"\\%03o", *(unsigned char*)(x+i)); j+= 4; }
  R[j++]= '"';
  R[j++]= 0;
  return R;
}

static void Kcreate_button(Uwidget_t *W, FILE *f)
{
  fprintf(f, "  widgets[%s]= gfx_button(%s, '%s', %s,GFXCLASS_BUTTON);\n",
        W->ident, Cstr(get_field(W,"text", NULL)),
                  get_field(W,"iconpos", "T"),
                  nustr(get_field(W,"icon",NULL))
  );
}

static void Kcreate_filelist(Uwidget_t *W,FILE *f)
{
  fprintf(f, "  widgets[%s]= gui_filelist_new(%s,%s,NULL,NULL,NULL);\n",
        W->ident, get_field(W,"types","1"), get_field(W,"hidden","1") 
  );
}

static void Kcreate_dirlist(Uwidget_t *W,FILE *f)
{
  fprintf(f, "  widgets[%s]= gui_dirlist_new(NULL,NULL,NULL);\n",
        W->ident 
  );
}

static void Kcreate_iconlist(Uwidget_t *W, FILE *f)
{
  fprintf(f, "  widgets[%s]= gui_iconlist_new(%s,%s,%s,%s,%s,%s);\n",
        W->ident, 
          get_field(W, "fg",   "0x00"),
          get_field(W, "bg",   "0xffFFff"),
          get_field(W, "vpad", "10"),
          get_field(W, "hpad", "10"),
          get_field(W, "max_text_width", "150"),
          get_field(W, "multi_select", "0")
  );
}


static void Kcreate_toggle(Uwidget_t *W, FILE *f)
{
  fprintf(f, "  widgets[%s]= gfx_button(%s, '%s', %s,GFXCLASS_TOGGLE);\n",
        W->ident, Cstr(get_field(W,"text", NULL)),
                  get_field(W,"iconpos", "T"),
                  nustr(get_field(W,"icon",NULL))
  );
}

static void Kcreate_label(Uwidget_t *W, FILE *f)
{
  fprintf(f, "  widgets[%s]= gfx_label(%s,%s,%s,'%s',%s);\n", 
             W->ident, 
    get_field(W,"font","1"),
    get_field(W,"fg", "0"),
    get_field(W,"bg", "1"),
    get_field(W,"just","C"),
    Cstr(get_field(W,"text",""))
  );
}

static void Kcreate_logview(Uwidget_t *W,FILE *f)
{
  fprintf(f,"  widgets[%s]= gui_logview();\n", W->ident);
}

static void Kcreate_radio(Uwidget_t *W, FILE *f)
{
  char *gr;
  gr= get_field(W,"group","-1");
  if (!strcmp(gr,"-1")) gr= W->ident;
  fprintf(f, "  widgets[%s]= radio_new(%s,%s);\n", 
                W->ident, Cstr(get_field(W,"text","")), gr);
}

static void Kcreate_checkbox(Uwidget_t *W, FILE *f)
{
  fprintf(f, "  widgets[%s]= gui_checkbox_new(%s,NULL,NULL);\n", 
                W->ident, Cstr(get_field(W,"text","")));
}

static void Kcreate_scrollbar(Uwidget_t *W,FILE *f)
{
  fprintf(f, "  widgets[%s]= scrollbar_new(%s,%s,%s,%s,1,NULL,NULL);\n",
               W->ident, get_field(W,"vertical","0"),
                         get_field(W,"init_start","0.0"),
                         get_field(W,"init_end","1.0"),
                         get_field(W,"init_step","0.1") );
}

static void Kcreate_progressbar(Uwidget_t *W, FILE *f)
{
  fprintf(f,"   widgets[%s]= gui_progressbar_new(%s,%s,%s,%s);\n",
                    W->ident, get_field(W,"font","1"),
                              get_field(W,"fg", "0x00"),
                              get_field(W,"bg", "0xededed"),
                              get_field(W,"fill", "0xcc"));
}

static void Kcreate_slider(Uwidget_t *W,FILE *f)
{
  fprintf(f, "  widgets[%s]= gui_slider_new(%s?'V':'H');\n",
               W->ident, get_field(W,"orientation","0"));
}

static void Kcreate_spinner(Uwidget_t *W,FILE *f)
{
  fprintf(f, "  widgets[%s]= gui_spinner('I',3);\n",
               W->ident);
}


static void Kcreate_panel(Uwidget_t *W, FILE *f)
{
  int i,n;
  Uwidget_t *cw;
  Uscreen_t *S;
  fprintf(f,"  {\n"
            "    panel_t *PP; \n"
            "    widgets[%s]= (widget_t*) (PP= panel_new(%d));\n",
               W->ident,W->n_subs); 
  for(i=0;i<W->n_subs;i++)
  {
    
    fprintf(f,"    static int WL_%d[]= {\n", i);
    S= W->subs[i];
    for(n=0,cw=S->widgets;cw;cw=cw->next,n++) fprintf(f,"    %s,\n", cw->ident);
    fprintf(f,"    };\n");
    fprintf(f,"    PP->subs[%d].n_widgets= %d;\n", i, n);
    fprintf(f,"    PP->subs[%d].widgets= WL_%d;\n", i,i);
  }
  fprintf(f,"  }\n");
}

static void Kcreate_dropdown(Uwidget_t *W,FILE *f)
{
  fprintf(f,"  widgets[%s]= dropdown_new(NULL,NULL);\n",W->ident);
}

static void Kcreate_list(Uwidget_t *W,FILE *f)
{
  fprintf(f,"  widgets[%s]= gui_list_new(0,0,NULL,NULL,NULL);\n",W->ident);
}

static void Kcreate_pixgrid(Uwidget_t *W,FILE *f)
{
  fprintf(f,"  widgets[%s]= gfx_pixgrid();\n",W->ident);
}

static void Kcreate_textentry(Uwidget_t *W,FILE *f)
{
  fprintf(f,"  widgets[%s]= text_entry_new(%s);\n",W->ident,
                       Cstr(get_field(W,"text","")) );
}

static void Kcreate_canvas(Uwidget_t *W,FILE *f)
{
   fprintf(f,"  widgets[%s]= gui_canvas_new(%s,%s);\n",
              W->ident,
              get_field(W,"horscroll", "1"),
              get_field(W,"verscroll", "1"));
}


static field_t fields_button[]= 
 {
   { "text", TYP_STR, NULL },
   { "iconpos", TYP_INT, "T" },
   { "icon", TYP_COLORIMG, NULL },
 };
 
static Uklass_t klass_button= { 
   "Button", 3, fields_button, 
   Kcreate_button, EDITFUNC(ew_button) };

static field_t fields_filelist[]= 
 {
   { "hidden", TYP_INT, "1" },
   { "types", TYP_INT, "1" },
 };

static Uklass_t klass_filelist= {
   "FileList", 2, fields_filelist, 
   Kcreate_filelist, EDITFUNC(ew_filelist) };

static Uklass_t klass_dirlist= {
   "DirList", 0, NULL,
   Kcreate_dirlist, EDITFUNC(ew_none) };

static field_t fields_iconlist[]= 
 {
  { "hpad", TYP_INT, "10" },
  { "vpad", TYP_INT, "10" },
  { "fg",   TYP_INT, "0x00" },
  { "bg",   TYP_INT, "0xffFFff" },
  { "max_text_width", TYP_INT, "150" },
  { "multi_select", TYP_INT, "0" }
 };

static Uklass_t klass_iconlist= {
   "IconList", 6, fields_iconlist,
   Kcreate_iconlist, EDITFUNC(ew_iconlist) };

static field_t fields_toggle[]=
 {
   { "text", TYP_STR, NULL },
   { "iconpos", TYP_INT, "T" },
   { "icon", TYP_COLORIMG, NULL },
 };
static Uklass_t klass_toggle= {
   "Toggle", 3, fields_toggle,
   Kcreate_toggle, EDITFUNC(ew_toggle) };

static field_t fields_label[]=
 {
   { "text", TYP_STR, NULL },
   { "fg",   TYP_INT, "0x000000" },
   { "bg",   TYP_INT, "0xededed" },
   { "font", TYP_INT, "1" },
   { "just", TYP_INT, "C" },
 };

static Uklass_t klass_label=
{
   "Label", 5, fields_label, Kcreate_label, EDITFUNC(ew_label)
};

static Uklass_t klass_logview= {
   "LogView", 0, NULL, Kcreate_logview, EDITFUNC(ew_none) };

static field_t fields_radio[]= 
 {
   { "text", TYP_STR, NULL },
   { "group", TYP_INT, "-1"},
 };

static Uklass_t klass_radio=
{
  "Radio", 2, fields_radio, Kcreate_radio, EDITFUNC(ew_radio)
};

static field_t fields_canvas[]=
{
  { "horscroll", TYP_INT, "1" },
  { "verscroll", TYP_INT, "1" },
};

static Uklass_t klass_canvas=
{
  "Canvas", 2, fields_canvas, Kcreate_canvas, EDITFUNC(ew_canvas)
};

static field_t fields_checkbox[]= 
 {
   { "text", TYP_STR, NULL },
 };

static Uklass_t klass_checkbox=
{
  "CheckBox", 1, fields_checkbox, Kcreate_checkbox, EDITFUNC(ew_checkbox)
};


static Uklass_t klass_panel =
{
  "Panel", 0, NULL, Kcreate_panel, EDITFUNC(ew_panel)
};

static field_t fields_progressbar[]=
 {
   { "fg",   TYP_INT, "0x000000" },
   { "bg",   TYP_INT, "0xededed" },
   { "fill", TYP_INT, "0x0000cc" },
   { "font", TYP_INT, "1" },
 };

static Uklass_t klass_progress=
{
   "ProgressBar", 4, fields_progressbar, Kcreate_progressbar, 
                        EDITFUNC(ew_progressbar)
};


static field_t fields_scrollbar[]=
 {
   { "vertical", TYP_INT, "0" },
   { "init_start", TYP_INT, "0.0" },
   { "init_end", TYP_INT, "1.0" },
   { "init_step", TYP_INT, "0.1" },
 };

static Uklass_t klass_scrollbar=
{
  "ScrollBar", 4, fields_scrollbar, Kcreate_scrollbar, EDITFUNC(ew_scrollbar)
};

static Uklass_t klass_dropdown=
 { "DropDown", 0, NULL , Kcreate_dropdown, EDITFUNC(ew_none) };

static Uklass_t klass_list=
 { "List", 0, NULL , Kcreate_list, EDITFUNC(ew_none) };

static Uklass_t klass_pixgrid=
 { "PixGrid", 0, NULL, Kcreate_pixgrid, EDITFUNC(ew_none) };

static field_t fields_textentry[]=
 {
   { "text", TYP_STR, NULL },
 };

static Uklass_t klass_textentry=
{
   "TextEntry", 1, fields_textentry, Kcreate_textentry, EDITFUNC(ew_textentry)
};

static field_t fields_slider[]=
{
   { "orientation", TYP_INT, "0" },
};
static Uklass_t klass_slider=
{
  "Slider", 1, fields_slider, Kcreate_slider, EDITFUNC(ew_slider)
};

static field_t fields_spinner[]= 
 {
   { "type", TYP_INT, "I" },
   { "decimals", TYP_INT, "2" },
 };

static Uklass_t klass_spinner= 
{
   "Spinner", 1, fields_spinner, Kcreate_spinner, EDITFUNC(ew_none)
};

Uklass_t *klasses[]=
{ &klass_button,
  &klass_canvas,
  &klass_checkbox,
  &klass_dropdown,
  &klass_filelist,
  &klass_dirlist,
  &klass_iconlist,
  &klass_label,
  &klass_list,
  &klass_logview,
  &klass_panel,
  &klass_pixgrid,
  &klass_progress,
  &klass_radio,
  &klass_scrollbar,
  &klass_slider,
  &klass_spinner,
  &klass_textentry,
  &klass_toggle,
  NULL
};


#ifndef KLASSES_NO_GFX
Uklass_t *choose_klass()
{
  Uklass_t *ret;
  gfx_event_t event;
  ret= NULL;
  gfx_SCRwidsel();
  while(1)
   if (gfx_event(NULL,&event))
   switch(event.type)
   {
   case WwselCANCEL:
   case GFXEV_QUIT:                             goto done;
   case WwselBUTTON:    ret= &klass_button;     goto done;
   case WwselCANVAS:    ret= &klass_canvas;     goto done;
   case WwselCHECKBOX:  ret= &klass_checkbox;   goto done;
   case WwselDROPDOWN:  ret= &klass_dropdown;   goto done;
   case WwselFILELIST:  ret= &klass_filelist;   goto done;
   case WwselDIRLIST:   ret= &klass_dirlist;    goto done;
   case WwselICONLIST:  ret= &klass_iconlist;   goto done;
   case WwselLABEL:     ret= &klass_label;      goto done;
   case WwselLIST:      ret= &klass_list;       goto done;
   case WwselLOGVIEW:   ret= &klass_logview;    goto done;
   case WwselPANEL:     ret= &klass_panel;      goto done;
   case WwselPIXGRID:   ret= &klass_pixgrid;    goto done;
   case WwselPROGRESS:  ret= &klass_progress;   goto done;
   case WwselRADIO:     ret= &klass_radio;      goto done;
   case WwselSCROLLBAR: ret= &klass_scrollbar;  goto done;
   case WwselSLIDER:    ret= &klass_slider;     goto done;
   case WwselSPINNER:   ret= &klass_spinner;    goto done;
   case WwselTEXTENTRY: ret= &klass_textentry;  goto done;
   case WwselTOGGLE:    ret= &klass_toggle;     goto done;
   }
done:
  gfx_leave_screen();
  return ret;
}
#endif
