#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "font_settings.h"


typedef struct strbuf {
   struct strbuf *prev,*next;
   char data[4076];
   int len;
} strbuf_t;


static void strbuf_puts_l(strbuf_t *sb,char *str,int len);
static  void strbuf_puts (strbuf_t * sb, char *str);
static  strbuf_t *strbuf_init ();
static  char *strbuf_collect (strbuf_t * sb);

static int sb_collect(strbuf_t *sb, char *dst) ;
static void sb_free(strbuf_t *sb) ;

static void strbuf_puts(strbuf_t *sb,char *str) {
	strbuf_puts_l(sb,str,strlen(str));
}

static void strbuf_puts_l(strbuf_t *sb,char *str,int len) {
	while(len) {
		strbuf_t *t;
		int cp;
		t= sb->prev;
		cp= sizeof(t->data)-t->len;
		if (cp>len) cp= len;
		if (cp==0) {
			
	t= (typeof(t)) calloc(1, sizeof(t[0]));


			t->len= 0;
			 {
	(t)->prev= (sb)->prev; 
	(t)->next= (sb); 
	(sb)->prev->next= (t); 
	(sb)->prev= (t); 
}


			continue;
		}
		memcpy(t->data+t->len, str, cp);
		str+= cp;
		len-= cp;
		t->len+= cp;
	}
}

static strbuf_t* strbuf_init() {
	strbuf_t *t;
	
	t= (typeof(t)) calloc(1, sizeof(t[0]));


	t->len= 0;
	t->prev= t->next= t;
	return t;
}


static char *strbuf_collect(strbuf_t *sb) {
	int l;
	char *k;
	l= sb_collect(sb, 0);
	k= malloc( l + 1);
	sb_collect(sb, k);
	k[l]= 0;
	sb_free(sb);
	return k;
}


static  int sb_collect(strbuf_t *sb, char *dst) {
	strbuf_t *k;
	int L;
	k= sb;
	L= 0;
	do {
		if (dst) 
			memcpy(dst+L, k->data, k->len);
		L+= k->len;
		k= k->next;
	} while(k!=sb);
	return L;
}

static void sb_free(strbuf_t *sb) 
{
	while(sb->prev!=sb){ 
		strbuf_t *t= sb->prev;
		sb->prev= t->prev;
		free(t);
	}
	free(sb);
}



static void sbint(strbuf_t *B,int V)
{
  char tb[40];
  sprintf(tb,"%d\n",V);
  strbuf_puts(B, tb);
}

void font_settings_to_str(font_settings_t *fs, char **buf)
{
  int i;
  strbuf_t *B;
  B= strbuf_init();
  sbint(B, fs->n_font_files);
  for(i=0;i<fs->n_font_files;i++)
  {
    strbuf_puts(B, fs->font_files[i]);
    strbuf_puts(B,"\n");
  }
  sbint(B, fs->n_fonts);
  for(i=0;i<fs->n_fonts;i++)
  {
    sbint(B, fs->fonts[i].font_file);
    sbint(B, fs->fonts[i].size);
  }
  *buf= strbuf_collect(B);
}

static char *sgetline(char **Rbuf)
{
  char *buf,*z;
again:
  buf= *Rbuf;
  z= strchr(buf,'\n');
  if (!z) { z= buf+strlen(buf); *Rbuf= z; }
     else { *Rbuf= z+1; *z= 0; }
  if (buf[0]=='#') goto again;
  return buf;
}

void font_settings_from_str(font_settings_t **Rfs, char *buf)
{
  int i;
  font_settings_t *fs;
  fs= malloc(sizeof(*fs));
  fs->n_font_files= atoi(sgetline(&buf));
  fs->font_files= malloc(sizeof(fs->font_files[0])*fs->n_font_files);
  for(i=0;i<fs->n_font_files;i++)
    fs->font_files[i]= strdup(sgetline(&buf));
  fs->n_fonts= atoi(sgetline(&buf));
  fs->fonts= malloc(sizeof(fs->fonts[0])*fs->n_fonts);
  for(i=0;i<fs->n_fonts;i++)
  {
    fs->fonts[i].font_file= atoi(sgetline(&buf));
    fs->fonts[i].size= atoi(sgetline(&buf));
  }
  *Rfs= fs;
}

char* generate_font_data(FILE *f, font_settings_t *fs)
{
  FILE *input;
  int i,j,R;
  unsigned char buf[100];
  char ebuf[2000];
  for(i=0;i<fs->n_font_files;i++)
  {
    input= fopen(fs->font_files[i],"r");
    if (!input) 
    {
       sprintf(ebuf,"can't open %s: %s\n", fs->font_files[i],
                   strerror(errno));
       return strdup(ebuf);
    }
    fprintf(f, "static const unsigned char fontdata%d[]={\n", i);
    while((R= fread(buf, 1, 18, input))>0)
    {
      for(j=0;j<R;j++) fprintf(f,"%d,",buf[j]);
      fprintf(f,"\n");
    }
    fprintf(f,"};\n");
    fclose(input);
  }
  fprintf(f,"#define NFONTS %d\n", fs->n_fonts);
  fprintf(f,"static fontdef_t fontdefs[NFONTS]= {\n");
  for(i=0;i<fs->n_fonts;i++)
    fprintf(f,"{ fontdata%d, %d },\n", fs->fonts[i].font_file, 
                                       fs->fonts[i].size);
  fprintf(f,"};\n");
  return NULL;
}
