
/***********************************************************************
 *                                                                     *
 *                         DIRECTORY LIST                              *
 * #W.DIRLIST                                                          *
 ***********************************************************************
  This is a simplified version of the file list. It can be used to 
  select directories. I didn't include any widgets other than a list,
  but maybe I could put some buttons and what not so that it becomes
  a builtin dialog.
 ***********************************************************************/

// #W.DIRLIST
typedef struct
{
  widget_t widget;

  char *dir;
  int nfiles;
  char **files;

  gui_list_t *w_list;

  void *evarg;
  event_function_t clickf,dlickf;
} gui_dirlist_t;



static void gui_dirlist_set_directory(gui_dirlist_t *FL,char *dir)
{
  int i;
  static const unsigned int *icons[3];
  static char fuu[]= "..";

  file_list_icons(icons);

  dir= path_absolute(dir);
  if (FL->dir) free(FL->dir);
  if (FL->files) free(FL->files);
  FL->nfiles= 0; FL->files= NULL; FL->dir= NULL;

  FL->dir= dir;
  FL->files= list_files(FL->dir);
  file_types(FL->dir, FL->files);

  for(i=0;FL->files[i];i++) ;
  FL->nfiles= i;

  FL->nfiles= filter_files_byspecial(FL->nfiles,FL->files);
  sort_files(FL->nfiles,FL->files);

  gfx_list_show(FL->w_list, 0);
  gfx_list_clear(FL->w_list);

  if (strcmp(FL->dir,"/"))
    gfx_list_insert(FL->w_list, -1, "<PARENT DIRECTORY>", fuu, icons[1]);
  for(i=0;i<FL->nfiles;i++)
  {
    gfx_list_insert( FL->w_list, -1, FL->files[i]+1, FL->files[i], 
           icons[ FL->files[i][0]=='F' ? 0: (FL->files[i][0]=='D'?1:2) ] );
  }
  gfx_list_show(FL->w_list,1);
}

static void gui_dirlist_show_all(gui_dirlist_t *FL,int show)
{
  widget_show(&FL->w_list->widget,show);
}

static void gui_dirlist_show(widget_t *W)
{
  gui_dirlist_t *FL;
  FL= (gui_dirlist_t*) W;
  if (W->shown)
    gfx_fill_rect(0, W->x, W->y, W->w, W->h, 0xededed);
  gui_dirlist_show_all(FL, 1);
  if (W->shown)
    gfx_update(W->x, W->y, W->w, W->h);
}
static void gui_dirlist_hide(widget_t *W)
{
  gui_dirlist_t *FL;
  FL= (gui_dirlist_t*) W;
  gui_dirlist_show_all(FL, 0);
}


static void gui_dirlist_position(widget_t *W, int x,int y,int w,int h)
{
  gui_dirlist_t *FL;
  int PAD;

// FIXME: convert these to resources
  PAD= 0;
  FL= (gui_dirlist_t*) W;

  W->x= x; W->y= y; W->w= w; W->h= h;
  if (W->shown) gfx_fill_rect(0, x,y,w,h, 0xededed);
  widget_move(&FL->w_list->widget,W->x+PAD, W->y+PAD, w-2*PAD, h-2*PAD);
  if (W->shown) gfx_update(x,y,w,h);
}

static int gui_dirlist_ipret_click(gui_dirlist_t *FL, gfx_event_t *event)
{
  int sel;
  char *f;
  char *d;

  sel= event->key;
  if (!FL->dir || sel>FL->nfiles || sel<0) return 1;

  if (strcmp(FL->dir,"/"))
  {
    if (sel==0) // go up
    {
      gui_dirlist_set_directory(FL,path_child(FL->dir, ".."));
      return 0;
    }
    sel--;
  }
  f= FL->files[sel];
  if (f[0]!='D') return 1; // not a directory, don't generate an event.
  
  gui_dirlist_set_directory(FL,d=path_child(FL->dir, f+1));
  free(d);
  return 0;
}

static int gui_dirlist_listclick(void *arg, gfx_event_t *event)
{
  gui_dirlist_t *FL;
  FL= arg;
  if (gui_dirlist_ipret_click(FL,event)) return 1;
  event->type= FL->widget.number;
  event->string= FL->dir;
  if (FL->clickf && (*FL->clickf)(FL->evarg,event)) return 1;
  return 0;
} 

static int gui_dirsel_listdlick(void *arg, gfx_event_t *event)
{
  gui_dirlist_t *FL;
  FL= arg;
  if (gui_dirlist_ipret_click(FL,event)) return 1;
  event->type= GFXEV(FL->widget.number, DOUBLE_CLICK);
  event->string= FL->dir;
  if (FL->dlickf && (*FL->dlickf)(FL->evarg,event)) return 1;
  return 0;
} 

static void gui_dirlist_destroy(widget_t *W)
{
  gui_dirlist_t *FL;
  FL= (gui_dirlist_t*) W;
  // gui_list_destroy(&FL->w_list->widget); FIXME: implement this
  freeNnull(FL->w_list); 
  freeNnull(FL->dir);
  freeNnull(FL->files);
}

static widget_t *gui_dirlist_new
  (void *evarg,event_function_t clickf, event_function_t dlickf)
{
  gui_dirlist_t *FL;
  FL= calloc(1,sizeof(*FL));

  FL->widget.klass= GFXCLASS_DIRLIST;
  FL->widget.show= gui_dirlist_show;
  FL->widget.hide= gui_dirlist_hide;
  FL->widget.destroy= gui_dirlist_destroy;
  FL->widget.position= gui_dirlist_position;

  FL->evarg= evarg;
  FL->clickf= clickf;
  FL->dlickf= dlickf;

  FL->w_list= (gui_list_t*) gui_list_new(0,0, FL, 
                               gui_dirlist_listclick, 
                               gui_dirsel_listdlick);
  return &FL->widget;
}

void gfx_dirlist_set_directory(int wno,char *dir)
{
  gui_dirlist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_DIRLIST);
  if (!FL) return ;
  gui_dirlist_set_directory(FL, dir);
}

char *gfx_dirlist_get_directory(int wno)
{
  gui_dirlist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_DIRLIST);
  if (!FL || !FL->dir) return strdup(".");
  return strdup(FL->dir);
}

void gfx_dirlist_refresh(int wno)
{
  gui_dirlist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_DIRLIST);
  if (!FL || !FL->dir) return ;
  gui_dirlist_set_directory(FL, FL->dir);
}

