#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ptrbuf.h"
#include "strbuf.h"
#include "expr.h"
#include "widget.h"
#include "gfx.h"
#include "edit_attr.h"
#include "show_resources.h"


static int show_text_field(field_t *F, char *name, int widget)
{
  if (strcmp(F->name,name)) return 0;
  gfx_set_text(widget, F->data);
  return 1;
}

static void show_field(field_t *F)
{
  if (show_text_field(F,"text",WGtext)
      || show_text_field(F,"group", WGgroup))
    return ;
  
  printf("unknown attrib: %s\n",F->name);
}

static void show_attribs(Uwidget_t *W)
{
  int i;
  field_t *F;
  for(i=0;i<W->klass->nfields;i++)
     show_field(W->klass->fields+i);
  for(F=W->fields;F;F=F->next)
     show_field(F);
}

static field_t *find_widget_field(Uwidget_t *W,char *name)
{
  field_t *F;
  for(F=W->fields;F;F=F->next)
    if (!strcmp(F->name,name)) return F;
  return NULL;
}
static int apply_value(Uwidget_t *W,field_t *ref, char *S)
{
  field_t *F;
  F= find_widget_field(W, ref->name);
  
  if (F)
  {
    F->data= S;
    return 0;
  }
  if (S[0])
  {
    F= malloc(sizeof(*F));
    F->name= strdup(ref->name);
    F->type= ref->type;
    F->data= S;
    F->next= W->fields;
    W->fields= F;
  }
  return 0;
}

static int apply_tfield(Uwidget_t *W, field_t *ref, int ent)
{
  char *S;
  S= gfx_get_text(ent);
  return apply_value(W, ref, S);
}

static void apply_selfield
  (Uwidget_t *W,field_t *ref, int selw,int nent, char **ent)
{
  int K;
  K= gfx_get_selected(selw);
  if (K<0 || K>=nent) return ;
  apply_value(W, ref, strdup(ent[K]));
}

static int apply_field(Uwidget_t *W, field_t *ref)
{
  char *N;
  N= ref->name;
  if (!strcmp(N, "text")) return apply_tfield(W,ref,WGtext);
  if (!strcmp(N, "group")) return apply_tfield(W,ref,WGgroup);
  return 0;
}

static void get_attribs(Uwidget_t *W)
{
  int i;
  for(i=0;i<W->klass->nfields;i++)
     apply_field(W,W->klass->fields+i);
}


void edit_attribs(Uwidget_t *W)
{
  char *sv;
  char buf[200];

  if (W->klass->edit)
  {
    (*W->klass->edit)(W);
    return ;
  }

  gfx_SCRediwid();

  snprintf(buf,200,"Widget: %s Class: %s",W->ident,W->klass->ident);
  gfx_set_text(WGident, buf);


  show_attribs(W);

  while(1)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case WGok: 
       get_attribs(W);
    case GFXEV_QUIT: 
    case WGcancel:
       goto done;
    }
  }
done:
  gfx_leave_screen();
}


