#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "pngio.h"


/***********************************************************************
 *                                                                     *
 *                  Imports a PNG image as a resource                  *
 *                                                                     *
 ***********************************************************************
 command line arguments:
  -g   output a graymap resource (default is color resource)
  -m   output a mask in addition to the data
 ***********************************************************************/

void usage(char **argv)
{
    exit(fprintf(stderr,"usage: %s [-g] file.png\n",argv[0]));
}

void shift(int *Ri, int *Rargc, char **argv, int n, ...)
{
  int i;
  int j;
  va_list ap;
  va_start(ap,n);
  if (n>1 && *Ri+n>*Rargc) usage(argv);
  for(i=*Ri,j=0;i+n<*Rargc;i++,j++)
  {
    if (j>0) *(va_arg(ap,char**)) = argv[i]; 
    argv[i]= argv[i+n];
  }
  *Ri--;
  *Rargc-= n;
  va_end(ap);
}

static int tlen= 0;

void putstart() { tlen= 0; }

void putint(unsigned int X)
{
    char buf[20];
    int lb;
    sprintf(buf,"%u,",X);
    lb= strlen(buf);
    if (tlen+lb>75) { printf("\n"); tlen= 0; }
    printf("%s", buf);
    tlen+= lb;   
}

unsigned int color_to_gray(unsigned int P)
{
  unsigned int R,G,B, V;
  R= (P>>16)&0xff;
  G= (P>> 8)&0xff;
  B= (P    )&0xff;

  V= R*0.3 + 0.6*G + 0.1*B;
  if (V>255) V= 255;
  return V;
}


int main(int argc,char **argv)
{
  char *fn;
  int i;
  unsigned int *data,w,h;
  char *e;
  int x,y;
  int mode= 'C';

  for(i=1;i<argc;i++)
  {
    if (!strcmp(argv[i],"-g")) 
       { mode= 'G'; shift(&i,&argc,argv,1); continue; }
  }

  if (argc!=2) usage(argv);

  fn= argv[1];
  e= read_png_image(fn, &w,&h, &data);
  if (e) exit(fprintf(stderr,"png error(%s): %s\n",fn,e));

  printf("{ .w= %d, .h= %d, .data= (unsigned %s[]){\n",
                 w,h, mode=='C' ? "int":"char");
  putstart();
  for(i=0;i<w*h;i++) putint(mode=='G' ? color_to_gray(data[i]) :data[i]);
  printf("}},\n");
  return 0;
}
