#include <stdio.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <stdlib.h>

struct my_error_mgr {
  struct jpeg_error_mgr pub;	
  jmp_buf setjmp_buffer;	
};

typedef struct my_error_mgr * my_error_ptr;



METHODDEF(void) my_error_exit (j_common_ptr cinfo) {
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  (*cinfo->err->output_message) (cinfo);
  longjmp(myerr->setjmp_buffer, 1);
}

int read_jpeg(char * filename,int *W,int *H,int **D) {
	int ret= 1;
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	FILE * infile;
	unsigned char* buffer[1];	
	int row_stride;

	int *d=0,w,h;
	buffer[0]= 0;

	if ((infile = fopen(filename, "rb")) == NULL) {
		fprintf(stderr, "can't open %s\n", filename);
		goto error_fopen;
	}

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	if (setjmp(jerr.setjmp_buffer))
		goto error_decompress;

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, infile);
	jpeg_read_header(&cinfo, TRUE);

	cinfo.out_color_space= JCS_RGB;
	cinfo.scale_num= 1;
	cinfo.scale_denom= 1;
	cinfo.dither_mode= JDITHER_NONE;

	jpeg_start_decompress(&cinfo);

	row_stride = cinfo.output_width * cinfo.output_components;
	buffer[0] = malloc(sizeof(buffer[0][0])*row_stride);

	w= cinfo.output_width;
	h= cinfo.output_height;
	d= malloc(sizeof(int)*w*h);
	if (!d)
		goto error_decompress;


	while (cinfo.output_scanline < cinfo.output_height) {
		jpeg_read_scanlines(&cinfo, buffer, 1);
		{
			int i;
			for(i=0;i<cinfo.output_width;i++)
				d[(cinfo.output_scanline-1)*w+i]=
					(buffer[0][i*3]<<16) |
					(buffer[0][i*3+1]<<8) |
					(buffer[0][i*3+2])
					;
		}
	}

	jpeg_finish_decompress(&cinfo);

	/* At this point you may want to check to see whether any corrupt-data
	* warnings occurred (test whether jerr.pub.num_warnings is nonzero).
	*/
	ret= 0;
	*D= d;
	*W= w;
	*H= h;
	d= 0;
error_decompress:
	if (d)
		free(d);
	if (buffer[0])
		free(buffer[0]);
	jpeg_destroy_decompress(&cinfo);
	fclose(infile);
error_fopen:
	return ret;
}


#ifdef rjpg_self_test

#include "wppm.c"

int main(int argc,char **argv) {
	int w,h,*d;
	if (argc!=3)
		exit(fprintf(stderr,"usage: %s input output\n",argv[0]));
	if (read_jpeg(argv[1],&w,&h,&d))	
		return 1;
	writeppm(argv[2],w,h,d);
	return 0;
}

#endif
